@@ -66,7 +66,7 @@ static BIO_METHOD * tlso_bio_setup( void );
 
 static int  tlso_opt_trace = 1;
 
-static void tlso_report_error( void );
+static void tlso_report_error( char *errmsg );
 
 static void tlso_info_cb( const SSL *ssl, int where, int ret );
 static int tlso_verify_cb( int ok, X509_STORE_CTX *ctx );
@@ -176,21 +176,23 @@ tlso_ca_list( char * bundle, char * dir, X509 *cert )
 	if ( bundle ) {
 		ca_list = SSL_load_client_CA_file( bundle );
 	}
-#if defined(HAVE_DIRENT_H) || defined(dirent)
 	if ( dir ) {
-		int freeit = 0;
+		char **dirs = ldap_str2charray( dir, CERTPATHSEP );
+		int freeit = 0, i, success = 0;
 
 		if ( !ca_list ) {
 			ca_list = sk_X509_NAME_new_null();
 			freeit = 1;
 		}
-		if ( !SSL_add_dir_cert_subjects_to_stack( ca_list, dir ) &&
-			freeit ) {
+		for ( i=0; dirs[i]; i++ ) {
+			success += SSL_add_dir_cert_subjects_to_stack( ca_list, dir );
+		}
+		if ( !success && freeit ) {
 			sk_X509_NAME_free( ca_list );
 			ca_list = NULL;
 		}
+		ldap_charray_free( dirs );
 	}
-#endif
 	if ( cert ) {
 		X509_NAME *xn = X509_get_subject_name( cert );
 		xn = X509_NAME_dup( xn );
@@ -370,7 +372,7 @@ tlso_ctx_cipher13( tlso_ctx *ctx, char *suites, char **oldsuites )
  * initialize a new TLS context
  */
 static int
-tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
+tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server, char *errmsg )
 {
 	tlso_ctx *ctx = (tlso_ctx *)lo->ldo_tls_ctx;
 	int i;
@@ -439,7 +441,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			Debug1( LDAP_DEBUG_ANY,
 				   "TLS: could not set TLSv1.3 cipher list %s.\n",
 				   lo->ldo_tls_ciphersuite );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 #endif
@@ -448,7 +450,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			Debug1( LDAP_DEBUG_ANY,
 				   "TLS: could not set cipher list %s.\n",
 				   lo->ldo_tls_ciphersuite );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 	}
@@ -458,7 +460,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		if ( !SSL_CTX_set_default_verify_paths( ctx ) ) {
 			Debug0( LDAP_DEBUG_ANY, "TLS: "
 				"could not use default certificate paths" );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 	} else {
@@ -470,19 +472,41 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			if ( !X509_STORE_add_cert( store, cert )) {
 				Debug0( LDAP_DEBUG_ANY, "TLS: "
 					"could not use CA certificate" );
-				tlso_report_error();
+				tlso_report_error( errmsg );
 				return -1;
 			}
 		}
-		if (( lt->lt_cacertfile || lt->lt_cacertdir ) && !SSL_CTX_load_verify_locations( ctx,
-				lt->lt_cacertfile, lt->lt_cacertdir ) )
-		{
-			Debug2( LDAP_DEBUG_ANY, "TLS: "
-				"could not load verify locations (file:`%s',dir:`%s').\n",
-				lo->ldo_tls_cacertfile ? lo->ldo_tls_cacertfile : "",
-				lo->ldo_tls_cacertdir ? lo->ldo_tls_cacertdir : "" );
-			tlso_report_error();
-			return -1;
+		if ( lt->lt_cacertfile || lt->lt_cacertdir ) {
+			char **dirs, *dummy = NULL;
+			if ( lt->lt_cacertdir ) {
+				dirs = ldap_str2charray( lt->lt_cacertdir, CERTPATHSEP );
+			} else {
+				dirs = &dummy;
+			}
+			/* Start with the first dir in path */
+			if ( !SSL_CTX_load_verify_locations( ctx,
+				lt->lt_cacertfile, dirs[0] ) )
+			{
+				Debug2( LDAP_DEBUG_ANY, "TLS: "
+					"could not load verify locations (file:`%s',dir:`%s').\n",
+					lo->ldo_tls_cacertfile ? lo->ldo_tls_cacertfile : "",
+					dirs[0] ? dirs[0] : "" );
+				tlso_report_error( errmsg );
+				if ( dirs != &dummy )
+					ldap_charray_free( dirs );
+				return -1;
+			}
+			/* Then additional dirs, if any */
+			if ( dirs != &dummy ) {
+				if ( dirs[1] ) {
+					int i;
+					X509_STORE *store = SSL_CTX_get_cert_store( ctx );
+					X509_LOOKUP *lookup = X509_STORE_add_lookup( store, X509_LOOKUP_hash_dir() );
+					for ( i=1; dirs[i]; i++ )
+						X509_LOOKUP_add_dir( lookup, dirs[i], X509_FILETYPE_PEM );
+				}
+				ldap_charray_free( dirs );
+			}
 		}
 
 		if ( is_server ) {
@@ -494,7 +518,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 					"could not load client CA list (file:`%s',dir:`%s').\n",
 					lo->ldo_tls_cacertfile ? lo->ldo_tls_cacertfile : "",
 					lo->ldo_tls_cacertdir ? lo->ldo_tls_cacertdir : "" );
-				tlso_report_error();
+				tlso_report_error( errmsg );
 				return -1;
 			}
 
@@ -511,7 +535,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		if ( !SSL_CTX_use_certificate( ctx, cert )) {
 			Debug0( LDAP_DEBUG_ANY,
 				"TLS: could not use certificate.\n" );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 		X509_free( cert );
@@ -522,7 +546,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		Debug1( LDAP_DEBUG_ANY,
 			"TLS: could not use certificate file `%s'.\n",
 			lo->ldo_tls_certfile );
-		tlso_report_error();
+		tlso_report_error( errmsg );
 		return -1;
 	}
 
@@ -535,7 +559,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		{
 			Debug0( LDAP_DEBUG_ANY,
 				"TLS: could not use private key.\n" );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 		EVP_PKEY_free( pkey );
@@ -547,7 +571,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 		Debug1( LDAP_DEBUG_ANY,
 			"TLS: could not use key file `%s'.\n",
 			lo->ldo_tls_keyfile );
-		tlso_report_error();
+		tlso_report_error( errmsg );
 		return -1;
 	}
 
@@ -565,14 +589,14 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			Debug1( LDAP_DEBUG_ANY,
 				"TLS: could not use DH parameters file `%s'.\n",
 				lo->ldo_tls_dhfile );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 		if (!( bio_params( bio, dh ))) {
 			Debug1( LDAP_DEBUG_ANY,
 				"TLS: could not read DH parameters file `%s'.\n",
 				lo->ldo_tls_dhfile );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			BIO_free( bio );
 			return -1;
 		}
@@ -596,7 +620,7 @@ tlso_ctx_init( struct ldapoptions *lo, struct ldaptls *lt, int is_server )
 			Debug1( LDAP_DEBUG_ANY,
 				"TLS: could not set EC name `%s'.\n",
 				lo->ldo_tls_ecname );
-			tlso_report_error();
+			tlso_report_error( errmsg );
 			return -1;
 		}
 	/*
@@ -1427,15 +1451,17 @@ tlso_verify_ok( int ok, X509_STORE_CTX *ctx )
 
 /* Inspired by ERR_print_errors in OpenSSL */
 static void
-tlso_report_error( void )
+tlso_report_error( char *errmsg )
 {
 	unsigned long l;
-	char buf[200];
+	char buf[ERRBUFSIZE];
 	const char *file;
 	int line;
 
 	while ( ( l = ERR_get_error_line( &file, &line ) ) != 0 ) {
-		ERR_error_string_n( l, buf, sizeof( buf ) );
+		ERR_error_string_n( l, buf, ERRBUFSIZE );
+		if ( !*errmsg )
+			strcpy(errmsg, buf );
 #ifdef HAVE_EBCDIC
 		if ( file ) {
 			file = LDAP_STRDUP( file );
